from pykotor.resource.formats.bwm.bwm_data import BWM, BWMType, BWMFace
from pykotor.resource.formats.bwm.io_bwm import BWMBinaryReader, BWMBinaryWriter
from pykotor.resource.formats.bwm.bwm_auto import write_bwm, read_bwm
